#
# This module is used to generate the header and pseudo-c-extension-module
# body for an extension module c-api library.
#
# See "Extending and Embedding the Python Interpeter",
#         "1.12 Providing a C API for an Extension Module"
#
# The purpose of a c-api pseudo-extension-module is to provide a platform
# independent mechanism for accessing c-api functions which are built in
# a shared library without polluting the global namespace.
#
# Towards this end, the c-api functions are compiled "static" but pointers
# to them are placed in a jump table.
#
# The header defines "proxy-macros" which provide a typechecked means of
# calling the API functions through the jump table.
#
# The header also provides an "import_module" macro which is used to initialize
# a pointer to the API jump table.  So, including the library header generates
# a single pointer of static data in the including file.  Calling the import
# macro fills in the pointer at runtime.
#
# The generated output includes header and an pseudo-c-extension body code.
# It is presumed that the API functions themselves already exist in external
# files which either have global linkage *or* are included directly into
# the pseudo-c-extension as extra_headers.
#

PROTO = """
static %(rval)s %(function)s %(proto)s;
"""

PSEUDO = """
#define %(function)s (*(%(rval)s (*) %(proto)s ) %(module)s_API[%(num)s])
"""

HDR_BODY = """
#ifndef _%(module)s
#define _%(module)s

%(headers)s

#ifdef __cplusplus
extern "C" {
#endif

/* Header file for %(module)s */

#if !defined(_%(module)s_MODULE)

static void **%(module)s_API;

#define import_libnumarray()                                               \\
      {                                                                    \\
        PyObject *module = PyImport_ImportModule("%(module)s");            \\
        if (module != NULL) {                                              \\
          PyObject *module_dict = PyModule_GetDict(module);                \\
          PyObject *c_api_object =                                         \\
                 PyDict_GetItemString(module_dict, "_C_API");              \\
          if (PyCObject_Check(c_api_object)) {                             \\
            %(module)s_API = (void **)PyCObject_AsVoidPtr(c_api_object);   \\
          }                                                                \\
        }                                                                  \\
      }
#endif

/* Macros defining components of function prototypes */

%(macros)s

#ifdef _%(module)s_MODULE
  /* This section is used when compiling %(module)s */

  static PyObject *_%(module)sError;
  
  %(protos)s
#else
  /* This section is used in modules that use %(module)s */
  %(pseudos)s
#endif

  /* Total number of C API pointers */
#define %(module)s_API_pointers %(num)s

#ifdef __cplusplus
}
#endif

#endif /* !defined(_%(module)s) */
"""

def genhdr(module, file, header_includes):
    functions = open(file).readlines()
    headers = mkhdrs(header_includes)
    macros, protos, pseudos = "", "", ""
    num = 0
    fps = []
    for l in functions:
        rval, function, proto = l.rstrip().split()
        fname = module + "_" + function
        protos += PROTO % locals()
        pseudos += PSEUDO % locals()
        fps.append(function)
        num += 1
    hdr = open(module +".h","w")
    print >>hdr, HDR_BODY % locals()
    return fps

MOD_BODY = """
#include <Python.h>

#define _%(module)s_MODULE
#include "%(module)s.h"

%(headers)s

static PyObject *_%(module)sError;

void *%(module)s_API[] = {
%(fps)s
};

static PyMethodDef _%(module)sMethods[] = {
    {NULL,      NULL}        /* Sentinel */
};

/* platform independent*/
#ifdef MS_WIN32
__declspec(dllexport)
#endif

void init%(module)s(void)
{
    PyObject *m = Py_InitModule("%(module)s", _%(module)sMethods);
    PyObject *c_api_object;

    _%(module)sError = PyErr_NewException("_%(module)s.error", NULL, NULL);

    /* Create a CObject containing the API pointer array's address */
    c_api_object = PyCObject_FromVoidPtr((void *)%(module)s_API, NULL);

    if (c_api_object != NULL) {
      /* Create a name for this object in the module's namespace */
      PyObject *d = PyModule_GetDict(m);

      PyDict_SetItemString(d, "_C_API", c_api_object);
      PyDict_SetItemString(d, "error", _%(module)sError);
      Py_DECREF(c_api_object);
    }
}

"""

def mkhdrs(hdr_list):
    headers = map(lambda x: '#include "%s"' % x, hdr_list)
    headers = "\n".join(headers)
    return headers

def genmod(module, fps, extra_headers):
    headers = mkhdrs(extra_headers)
    fps  = "\t(void*)" + ",\n\t(void*)".join(fps)
    mod = open(module +".c","w")
    print >>mod, MOD_BODY % locals()

def main(module, functions, header_includes, body_includes):
    fps = genhdr(module, functions, header_includes)
    genmod(module, fps, body_includes)

if __name__ == "__main__":
    import sys

    module = sys.argv[1]
    functions = sys.argv[2]

    try:
        body = sys.argv.index("--body")
        header_includes = sys.argv[3:body]
        body_includes = sys.argv[body+1:]
    except:
        header_includes = []
        body_includes = sys.argv[3:]

    print "header_includes", header_includes
    print "body_includes", body_includes
    main(module, functions, header_includes, body_includes);
