/* $Id: buffer.ch,v 1.3 2002/03/11 20:05:50 jaytmiller Exp $ */

#include <Python.h>
#include <stdio.h>

/**********************************************************************/
/*  Buffer Utility Functions                                          */
/**********************************************************************/

static PyObject *getBuffer( PyObject *obj) {
  if (obj->ob_type->tp_as_buffer == NULL)
    return PyObject_CallMethod(obj, "__buffer__", NULL);
  Py_INCREF(obj);  /* Since CallMethod returns a new object when it
                      succeeds, We'll need to DECREF later to free it.
                      INCREF ordinary buffers here so we don't have to
                      remember where the buffer came from at DECREF time.
		   */
  return obj;
}

static int isBuffer (PyObject *obj) {
  /* Either it defines the buffer API, or it is an instance which returns
     a buffer when obj.__buffer__() is called */
  PyObject *buf = getBuffer(obj);
  Py_DECREF(buf);
  return buf->ob_type->tp_as_buffer != NULL;
}

/**********************************************************************/

static int getWriteBufferDataPtr(PyObject *buffobj, void **buff) {
  int rval = -1;
  if ((buffobj = getBuffer(buffobj)) &&
      buffobj->ob_type->tp_as_buffer->bf_getwritebuffer)
    {
      rval = buffobj->ob_type->tp_as_buffer->bf_getwritebuffer(buffobj, 
							       0, buff);
      Py_DECREF(buffobj);
    }
  return rval;
}

/**********************************************************************/

static int isBufferWriteable (PyObject *buffobj) {
  void *ptr;
  int rval = -1;
  rval = getWriteBufferDataPtr(buffobj, &ptr);
  if (rval == -1)
    PyErr_Clear(); /* Since we're just "testing", it's not really an error */
  return rval != -1;
}

/**********************************************************************/

static int getReadBufferDataPtr(PyObject *buffobj, void **buff) {
  int rval = -1;
  if ((buffobj = getBuffer(buffobj)) &&
      buffobj->ob_type->tp_as_buffer->bf_getreadbuffer)
    {
      rval = buffobj->ob_type->tp_as_buffer->bf_getreadbuffer(buffobj, 
							      0, buff);
      Py_DECREF(buffobj);
    }
  return rval;
}

/**********************************************************************/

static int getBufferSize(PyObject *buffobj) {
  int segcount, size=0;
  if ((buffobj = getBuffer(buffobj))) 
    {
      segcount = buffobj->ob_type->tp_as_buffer->bf_getsegcount(buffobj, 
								&size);
      Py_DECREF(buffobj);
    }
  else
    size = -1;
  return size;
}

static int NA_getBufferPtrAndSize(PyObject *buffobj, int readonly, void **ptr)
{
	int rval;
	if (readonly)
		rval = getReadBufferDataPtr(buffobj, ptr);
	else 
		rval = getWriteBufferDataPtr(buffobj, ptr);
	return rval;
}

static int NA_checkOneCBuffer(char *name, long niter, 
		     void *buffer, long bsize, size_t typesize)
{
	if (niter*typesize > bsize) {
		PyErr_Format(_libnumarrayError, 
			     "%s: access out of buffer. niter=%d typesize=%d bsize=%d",
			     name, (int) niter, (int) typesize, (int) bsize);
		return -1;
	}
	if (((long) buffer) % typesize) {
		PyErr_Format(_libnumarrayError, 
			     "%s: buffer not aligned on %d byte boundary.",
			     name, typesize);
		return -1;
	}
	return 0;
}

static int NA_checkIo(char *name, 
		      int wantIn, int wantOut, int gotIn, int gotOut)
{
	if (wantIn != gotIn) {
		PyErr_Format(_libnumarrayError,
		   "%s: wrong # of input buffers. Expected %d.  Got %d.", 
		    name, wantIn, gotIn);
		return -1;
	}
	if (wantOut != gotOut) {
		PyErr_Format(_libnumarrayError,
		   "%s: wrong # of output buffers. Expected %d.  Got %d.", 
		    name, wantOut, gotOut);
		return -1;
	}
	return 0;
}

static int NA_checkNCBuffers(char *name, int N, long niter, 
			    void **buffers, long *bsizes,
			    Int8 *typesizes, Int8 *iters)
{
	int i;
	for (i=0; i<N; i++)
		if (NA_checkOneCBuffer(name, iters[i] ? iters[i] : niter, 
			      buffers[i], bsizes[i], typesizes[i]))
			return -1;
	return 0;
}

#define MIN(x,y) (((x)<=(y)) ? (x) : (y))
#define MAX(x,y) (((x)>=(y)) ? (x) : (y))
#define ABS(x) (((x) >= 0) ? (x) : -(x))

static int NA_checkOneStriding(char *name, long dim, long *shape,
	       long offset, long *stride, long buffersize, int itemsize, 
	       int align)
{
  int i;
  long omin=offset, omax=offset;
  
  if (align && (offset % itemsize)) {
    PyErr_Format(_libnumarrayError, 
		 "%s: buffer not aligned on %d byte boundary.",
		 name, itemsize);
    return -1;
  }
  for(i=0; i<dim; i++) {
    long strideN = stride[i] * (shape[i]-1);
    long tmax = omax + strideN;
    long tmin = omin + strideN;
    if (shape[i]-1 >= 0) {  /* Skip dimension == 0. */
      omax = MAX(omax, tmax);
      omin = MIN(omin, tmin);
      if (align && (ABS(stride[i]) % itemsize)) {
	PyErr_Format(_libnumarrayError, 
         "%s: stride %d not aligned on %d byte boundary.",
	     name, (int) stride[i], itemsize);
	return -1;
      }
      if (omax + itemsize > buffersize) {
	PyErr_Format(_libnumarrayError, 
           "%s: access beyond buffer. offset=%d buffersize=%d",
		     name, (int) omax, (int) buffersize);
	return -1;
      }
      if (omin < 0) {
	PyErr_Format(_libnumarrayError, 
	     "%s: access before buffer. offset=%d buffersize=%d",
		     name, (int) omin, (int) buffersize);
	return -1;
      }
    }
  }
  return 0;
}
