#include <Python.h>

staticforward PyTypeObject CfuncType;

static void
cfunc_dealloc(PyObject* self)
{
    PyObject_Del(self);
}

static PyObject *
cfunc_repr(PyObject *self)
{
	char buf[256];
	CfuncObject *me = (CfuncObject *) self;
	sprintf(buf, "<cfunc '%s' at %08lx check-self:%d align:%d  io:(%d, %d)>", 
		 me->descr.name, (unsigned long ) me->descr.fptr, 
		 me->descr.chkself, me->descr.align, 
		 me->descr.wantIn, me->descr.wantOut);
	return PyString_FromString(buf);
}

/* Call a standard "stride" function
**
** Stride functions always take one input and one output array.
** They can handle n-dimensional data with arbitrary strides (of
** either sign) for both the input and output arrays. Typically
** these functions are used to copy data, byteswap, or align data.
**
**
** It expects the following arguments:
**
**   Number of iterations for each dimension as a tuple
**   Input Buffer Object
**   Offset in bytes for input buffer
**   Input strides (in bytes) for each dimension as a tuple
**   Output Buffer Object
**   Offset in bytes for output buffer
**   Output strides (in bytes) for each dimension as a tuple
**   An integer (Optional), typically the number of bytes to copy per
*       element.
**
** Returns None
**
** The arguments expected by the standard stride functions that this
** function calls are:
**
**   Number of dimensions to iterate over
**   Long int value (from the optional last argument to
**      callStrideConvCFunc)
**      often unused by the C Function
**   An array of long ints. Each is the number of iterations for each
**      dimension. NOTE: the previous argument as well as the stride
**      arguments are reversed in order with respect to how they are
**      used in Python. Fastest changing dimension is the first element
**      in the arrays!
**   A void pointer to the input data buffer.
**   The starting offset for the input data buffer in bytes (long int).
**   An array of long int input strides (in bytes) [reversed as with
**      the iteration array]
**   A void pointer to the output data buffer.
**   The starting offset for the output data buffer in bytes (long int).
**   An array of long int output strides (in bytes) [also reversed]
*/

static PyObject *
callStrideConvCFunc(PyObject *self, PyObject *args) {
    PyObject *inbuffObj, *outbuffObj, *shapeObj;
    PyObject *inbstridesObj, *outbstridesObj, *otemp;

    long nshape, ninbstrides, noutbstrides, nargs, i, nbytes = 0;
    long shape[MAXDIM], inbstrides[MAXDIM], outbstrides[MAXDIM];
    long inbsize, outbsize, inboffset, outboffset;

    CfuncObject *me = (CfuncObject *) self;
    CFUNC_STRIDE_CONV_FUNC funcptr;
    void *inbuffer, *outbuffer;

    nargs = PyObject_Length(args);
    if (!PyArg_ParseTuple(args, "OOlOOlO|l",
            &shapeObj, &inbuffObj,  &inboffset, &inbstridesObj,
            &outbuffObj, &outboffset, &outbstridesObj,
            &nbytes)) {
        return PyErr_Format(_libnumarrayError,
                 "%s: Problem with argument list",
		  me->descr.name);
    }
    /* Check that iteration and stride tuples have the same # elements */
    nshape = PyObject_Length(shapeObj);
    ninbstrides = PyObject_Length(inbstridesObj);
    noutbstrides = PyObject_Length(outbstridesObj);
    if ((nshape != ninbstrides) || (nshape != noutbstrides)) {
        return PyErr_Format(_libnumarrayError,
            "%s: Missmatch between iteration and strides tuples",
	    me->descr.name);
    }
    /* Copy values from Python tuples to simple C arrays (and reverse order) */
    for (i=nshape-1; i>=0; i--) {
        otemp = PySequence_GetItem(shapeObj, i);
        shape[nshape-i-1] = PyInt_AsLong(otemp);
        Py_DECREF(otemp);
        otemp = PySequence_GetItem(inbstridesObj, i);
        inbstrides[nshape-i-1] = PyInt_AsLong(otemp);
        Py_DECREF(otemp);
        otemp = PySequence_GetItem(outbstridesObj, i);
        outbstrides[nshape-i-1] = PyInt_AsLong(otemp);
        Py_DECREF(otemp);
    }

    if ((inbsize = NA_getBufferPtrAndSize(inbuffObj, 1, &inbuffer)) < 0)
	return PyErr_Format(_libnumarrayError,
            "%s: Problem with input buffer", me->descr.name);

    if ((outbsize = NA_getBufferPtrAndSize(outbuffObj, 0, &outbuffer)) < 0)
	return PyErr_Format(_libnumarrayError,
	    "%s: Problem with output buffer (read only?)", me->descr.name);

    /* Check buffer alignment and bounds */
    if (NA_checkOneStriding(me->descr.name, nshape, shape,
			    inboffset, inbstrides, inbsize,
			    (me->descr.sizes[0] == -1) ? 
			    nbytes : me->descr.sizes[0],
			    me->descr.align) ||
	NA_checkOneStriding(me->descr.name, nshape, shape,
			    outboffset, outbstrides, outbsize,
			    (me->descr.sizes[1] == -1) ? 
			    nbytes : me->descr.sizes[1],
			    me->descr.align))
	    return NULL;
    
    /* Cast function pointer and perform stride operation */
    funcptr = (CFUNC_STRIDE_CONV_FUNC) me->descr.fptr;
    if ((*funcptr)(nshape-1, nbytes, shape, 
		   inbuffer,  inboffset, inbstrides,
		   outbuffer, outboffset, outbstrides) == 0) {
	    Py_INCREF(Py_None);
	    return Py_None;
    } else {
	    return NULL;
    }
}

/* Convert a standard C numeric value to a Python numeric value.
**
** Handles both nonaligned and/or byteswapped C data.
**
** Input arguments are:
**
**   Buffer object that contains the C numeric value.
**   Offset (in bytes) into the buffer that the data is located at.
**   The size of the C numeric data item in bytes.
**   Flag indicating if the C data is byteswapped from the processor's
**     natural representation.
**
**   Returns a Python numeric value.
*/

static PyObject *
NumTypeAsPyValue(PyObject *self, PyObject *args) {
    PyObject *bufferObj;
    long offset, itemsize, byteswap, i, buffersize;
    double temp;  /* to hold copies of largest possible type */
    void *buffer;
    char *tempptr;
    CFUNCasPyValue funcptr;
    CfuncObject *me = (CfuncObject *) self;

    if (!PyArg_ParseTuple(args, "Olll", 
			  &bufferObj, &offset, &itemsize, &byteswap))
        return PyErr_Format(_libnumarrayError,
		"NumTypeAsPyValue: Problem with argument list");

    if ((buffersize = NA_getBufferPtrAndSize(bufferObj, 1, &buffer)) < 0)
        return PyErr_Format(_libnumarrayError,
                "NumTypeAsPyValue: Problem with array buffer");

    if (offset < 0)
	return PyErr_Format(_libnumarrayError,
		"NumTypeAsPyValue: invalid negative offset: %d", (int) offset);

    /* Guarantee valid buffer pointer */
    if (offset+itemsize > buffersize)
	    return PyErr_Format(_libnumarrayError,
		"NumTypeAsPyValue: buffer too small for offset and itemsize.");

    /* Do byteswapping.  Guarantee double alignment by using temp. */
    tempptr = (char *) &temp;
    if (!byteswap) {
        for (i=0; i<itemsize; i++)
            *(tempptr++) = *(((char *) buffer)+offset+i);
    } else {
        tempptr += itemsize-1;
        for (i=0; i<itemsize; i++)
            *(tempptr--) = *(((char *) buffer)+offset+i);
    }

    funcptr = (CFUNCasPyValue) me->descr.fptr;

    /* Call function to build PyObject.  Bad parameters to this function
       may render call meaningless, but "temp" guarantees that its safe.  */
    return (*funcptr)((void *)(&temp));
}

/* Convert a Python numeric value to a standard C numeric value.
**
** Handles both nonaligned and/or byteswapped C data.
**
** Input arguments are:
**
**   The Python numeric value to be converted.
**   Buffer object to contain the C numeric value.
**   Offset (in bytes) into the buffer that the data is to be copied to.
**   The size of the C numeric data item in bytes.
**   Flag indicating if the C data is byteswapped from the processor's
**     natural representation.
**
**   Returns None
*/

static PyObject *
NumTypeFromPyValue(PyObject *self, PyObject *args) {
    PyObject *bufferObj, *valueObj;
    long offset, itemsize, byteswap, i, buffersize;
    double temp;
    void *buffer;
    char *tempptr;
    CFUNCfromPyValue funcptr;
    CfuncObject *me = (CfuncObject *) self;

    if (!PyArg_ParseTuple(args, "OOlll", 
		  &valueObj, &bufferObj, &offset, &itemsize, &byteswap)) 
        return PyErr_Format(_libnumarrayError,
                 "%s: Problem with argument list", me->descr.name);

    if ((buffersize = NA_getBufferPtrAndSize(bufferObj, 0, &buffer)) < 0)
	    return PyErr_Format(_libnumarrayError,
                "%s: Problem with array buffer (read only?)", me->descr.name);

    funcptr = (CFUNCfromPyValue) me->descr.fptr;

    /* Convert python object into "temp". Always safe. */
    if (!((*funcptr)(valueObj, (void *)( &temp))))
        return PyErr_Format(_libnumarrayError,
		 "%s: Problem converting value", me->descr.name);

    /* Check buffer offset. */
    if (offset < 0)
	return PyErr_Format(_libnumarrayError,
		"%s: invalid negative offset: %ld", me->descr.name, offset);

    if (offset+itemsize > buffersize)
	return PyErr_Format(_libnumarrayError,
		"%s: buffer too small(%ld) for offset(%ld) and itemsize(%ld)",
			me->descr.name, buffersize, offset, itemsize);

    /* Copy "temp" to array buffer. */
    tempptr = (char *) &temp;
    if (!byteswap) {
        for (i=0; i<itemsize; i++)
            *(((char *) buffer)+offset+i) = *(tempptr++);
    } else {
        tempptr += itemsize-1;
        for (i=0; i<itemsize; i++)
            *(((char *) buffer)+offset+i) = *(tempptr--);
    }
    Py_INCREF(Py_None);
    return Py_None;
}

/* Function to call standard C Ufuncs
**
** The C Ufuncs expect contiguous 1-d data arrays, input and output arrays
** iterate with standard increments of one data element over all arrays.
** (There are some exceptions like arrayrangexxx which use one or more of
** the data arrays as parameter or other sources of information and do not
** iterate over every buffer).
**
** Arguments:
**
**   Number of iterations (simple integer value).
**   Number of input arrays.
**   Number of output arrays.
**   Tuple of tuples, one tuple per input/output array. Each of these
**     tuples consists of a buffer object and a byte offset to start.
**
** Returns None
*/

static PyObject *
callCUFunc(PyObject *self, PyObject *args) {
  PyObject *BufferObj, *DataArgs, *ArgTuple;
  long pnargs, ninargs, noutargs, niter, offset, i;
  UFUNC ufuncptr;
  CfuncObject *me = (CfuncObject *) self;
  char *buffers[MAXARGS];
  long bsizes[MAXARGS];

  if (!PyArg_ParseTuple(args, "lllO",
			&niter, &ninargs, &noutargs, &DataArgs))
     return PyErr_Format(_libnumarrayError,
         "%s: Problem with argument list", me->descr.name);

  /* check consistency of stated inputs/outputs and supplied buffers */
  pnargs = PyObject_Length(DataArgs);
  if ((pnargs != (ninargs+noutargs)) || (pnargs > MAXARGS)) 
     return PyErr_Format(_libnumarrayError,
       "%s: wrong buffer count for function", me->descr.name);

  /* Unpack buffers and offsets, get data pointers */
  for (i=0; i<pnargs; i++) {
     int readonly = (i < ninargs);
     ArgTuple = PySequence_GetItem(DataArgs, i);
     Py_DECREF(ArgTuple);
     if (!PyArg_ParseTuple(ArgTuple, "Ol", &BufferObj, &offset))
	     return PyErr_Format(_libnumarrayError,
		 "%s: Problem with buffer/offset tuple", me->descr.name);
     if (offset < 0)
	return PyErr_Format(_libnumarrayError,
	    "%s: invalid negative offset:%d for buffer[%d]", 
			me->descr.name, (int) offset, (int) i);
     if ((bsizes[i] = NA_getBufferPtrAndSize(BufferObj, readonly, 
					  (void *) &buffers[i])) < 0)
	 return PyErr_Format(_libnumarrayError,
	    "%s: Problem with %s buffer[%d].", 
			me->descr.name, readonly ? "read" : "write", (int) i);
     buffers[i] += offset;
     bsizes[i]  -= offset; /* "degrade" buffer size by offset. */
  }

  ufuncptr = (UFUNC) me->descr.fptr;

  /* If it's not a self-checking ufunc, check arg count match,
     buffer size, and alignment for all buffers */
  if (!me->descr.chkself && 
      (NA_checkIo(me->descr.name, 
		  me->descr.wantIn, me->descr.wantOut, ninargs, noutargs) ||
       NA_checkNCBuffers(me->descr.name, pnargs, 
			niter, (void **) buffers, bsizes, 
			me->descr.sizes, me->descr.iters)))
	  return NULL;

  /* Since the parameters are valid, call the C Ufunc */
  if (!(*ufuncptr)(niter, ninargs, noutargs, (void **)buffers, bsizes)) {
	  Py_INCREF(Py_None);
	  return Py_None;
  } else {
	  return NULL;
  }
}


/* Handle "calling" the cfunc object at the python level. 
   Dispatch the call to the appropriate python-c wrapper based
   on the cfunc type.  Do this dispatch to avoid having to
   check that python code didn't somehow create a mismatch between
   cfunc and wrapper.
*/
static PyObject *
cfunc_call(PyObject *self, PyObject *argsTuple, PyObject *argsDict)
{
	CfuncObject *me = (CfuncObject *) self;
	switch(me->descr.type) {
	case CFUNC_UFUNC:
		return callCUFunc(self, argsTuple);
		break;
	case CFUNC_STRIDING:
		return callStrideConvCFunc(self, argsTuple);
		break;
	case CFUNC_FROM_PY_VALUE:
		return NumTypeFromPyValue(self, argsTuple);
		break;
	case CFUNC_AS_PY_VALUE:
		return NumTypeAsPyValue(self, argsTuple);
		break;
	default:
		return PyErr_Format( _libnumarrayError,
		     "cfunc_call: Can't dispatch cfunc '%s' with type: %d.", 
		     me->descr.name, me->descr.type);
	}
}

static PyTypeObject CfuncType = {
    PyObject_HEAD_INIT(NULL)
    0,
    "Cfunc",
    sizeof(CfuncObject),
    0,
    cfunc_dealloc, /*tp_dealloc*/
    0,          /*tp_print*/
    0,          /*tp_getattr*/
    0,          /*tp_setattr*/
    0,          /*tp_compare*/
    cfunc_repr, /*tp_repr*/
    0,          /*tp_as_number*/
    0,          /*tp_as_sequence*/
    0,          /*tp_as_mapping*/
    0,          /*tp_hash */
    cfunc_call, /* tp_call */
};


/* CfuncObjects are created at the c-level only.  They ensure that each
cfunc is called via the correct python-c-wrapper as defined by its 
CfuncDescriptor.  The wrapper, in turn, does conversions and buffer size
and alignment checking.  Allowing these to be created at the python level
would enable them to be created *wrong* at the python level, and thereby
enable python code to *crash* python. 
*/ 
static PyObject*
NA_new_cfunc(CfuncDescriptor *cfd)
{
    CfuncObject* cfunc;
    
    CfuncType.ob_type = &PyType_Type;  /* Should be done once at init.
					  Do now since there is no init. */

    cfunc = PyObject_New(CfuncObject, &CfuncType);
    
    if (!cfunc) {
	    return PyErr_Format(_libnumarrayError,
			       "NA_new_cfunc: failed creating '%s'",
			       cfd->name);
    }

    cfunc->descr = *cfd;

    return (PyObject*)cfunc;
}

static int NA_add_cfunc(PyObject *dict, char *keystr, CfuncDescriptor *descr) {
    return PyDict_SetItemString(dict, keystr, 
				(PyObject *) NA_new_cfunc(descr));
}
