/* This file contains getNDInfo and getNumInfo.  Both functions assign
   values to a struct that describes a Python NDArray or NumArray object.
   The functions return 0 if all information was successfully obtained
   and assigned to the struct, and they return -1 on error.

   getNDInfo gets generic NDArray attributes.
   getNumInfo gets NumArray-specific attributes as well as the ones
   defined for NDArray.

   The calling sequence is the same for both:

   PyObject *object       the Python NDArray object for which info is needed
   NDInfo *NumArrayInfo   the C struct to be updated with info about object
*/

static int getNDInfo (PyObject *object, NDInfo *NumArrayInfo) {

	PyObject *temp, *item;
	void *buf;
	long length;
	long i;

	if (PyObject_HasAttrString (object, "_data")) {
	    temp = PyObject_GetAttrString (object, "_data");
	    if (getReadBufferDataPtr (temp, &buf) < 0)
		return (-1);
	    NumArrayInfo->data = buf;
	    NumArrayInfo->writeable = isBufferWriteable (temp);
	    Py_DECREF (temp);
	} else {
	    return (-1);
	}
	NumArrayInfo->imag = NULL;

	if (PyObject_HasAttrString (object, "_shape")) {
	    temp = PyObject_GetAttrString (object, "_shape");
	    length = PyObject_Length (temp);
	    NumArrayInfo->ndim = length;

	    for (i = 0;  i < length;  i++) {
		item = PySequence_GetItem (temp, i);
		NumArrayInfo->shape[i] = PyInt_AsLong (item);
		Py_DECREF (item);
	    }
	    for (i = length;  i < MAXDIM;  i++) {
		NumArrayInfo->shape[i] = 0;
	    }
	    Py_DECREF (temp);
	} else {
	    return (-1);
	}

	if (NumArrayInfo->ndim > 0) {
	    NumArrayInfo->nelem = 1;
	    for (i = 0;  i < NumArrayInfo->ndim;  i++)
		NumArrayInfo->nelem *= NumArrayInfo->shape[i];
	} else {
	    NumArrayInfo->nelem = 0;
	}

	if (PyObject_HasAttrString (object, "_strides")) {
	    temp = PyObject_GetAttrString (object, "_strides");
	    for (i = 0;  i < length;  i++) {
		item = PySequence_GetItem (temp, i);
		NumArrayInfo->strides[i] = PyInt_AsLong (item);
		Py_DECREF (item);
	    }
	    for (i = length;  i < MAXDIM;  i++) {
		NumArrayInfo->strides[i] = 0;
	    }
	    Py_DECREF (temp);
	} else {
	    return (-1);
	}

	if (PyObject_HasAttrString (object, "_byteoffset")) {
	    temp = PyObject_GetAttrString (object, "_byteoffset");
	    NumArrayInfo->byteoffset = PyInt_AsLong (temp);
	    Py_DECREF (temp);
	} else {
	    return (-1);
	}

	if (PyObject_HasAttrString (object, "_bytestride")) {
	    temp = PyObject_GetAttrString (object, "_bytestride");
	    NumArrayInfo->bytestride = PyInt_AsLong (temp);
	    Py_DECREF (temp);
	} else {
	    return (-1);
	}

	if (PyObject_HasAttrString (object, "_itemsize")) {
	    temp = PyObject_GetAttrString (object, "_itemsize");
	    NumArrayInfo->itemsize = PyInt_AsLong (temp);
	    Py_DECREF (temp);
	} else {
	    return (-1);
	}

	if (PyObject_HasAttrString (object, "_aligned")) {
	    temp = PyObject_GetAttrString (object, "_aligned");
	    NumArrayInfo->aligned = PyInt_AsLong (temp);
	    Py_DECREF (temp);
	} else {
	    return (-1);
	}

	if (PyObject_HasAttrString (object, "_contiguous")) {
	    temp = PyObject_GetAttrString (object, "_contiguous");
	    NumArrayInfo->contiguous = PyInt_AsLong (temp);
	    Py_DECREF (temp);
	} else {
	    return (-1);
	}

	/* This flag may be reassigned by other functions when more
	   attributes are available.
	*/
	NumArrayInfo->C_array =
	   (NumArrayInfo->aligned &&
	    NumArrayInfo->contiguous);

	return (0);
}

/* This routine gets the _type and _byteswap attributes, which are in the
   NumArray subclass, and it also gets the generic NDArray info.
*/

static int getNumInfo (PyObject *object, NDInfo *NumArrayInfo) {

	PyObject *temp, *item;

	/* Get generic NDArray attributes. */
	if (getNDInfo (object, NumArrayInfo))
	    return (-1);

	if (PyObject_HasAttrString (object, "_type")) {
	    temp = PyObject_GetAttrString (object, "_type");
	    if (PyObject_HasAttrString (temp, "name")) {
		item = PyObject_GetAttrString (temp, "name");
		strcpy (NumArrayInfo->type, PyString_AsString (item));
		Py_DECREF (item);
	    } else {
		return (-1);
	    }
	    Py_DECREF (temp);
	} else {
	    return (-1);
	}

	if (PyObject_HasAttrString (object, "_byteswap")) {
	    temp = PyObject_GetAttrString (object, "_byteswap");
	    NumArrayInfo->byteswap = PyInt_AsLong (temp);
	    Py_DECREF (temp);
	} else {
	    return (-1);
	}

	/* The C_array flag was set in getNDInfo, but we didn't have
	   byteswap at that time, so we need to update C_array.
	*/
	NumArrayInfo->C_array =
	   (NumArrayInfo->aligned &&
	    NumArrayInfo->contiguous &&
	   !NumArrayInfo->byteswap);
	if (NumArrayInfo->imag != NULL)
	    NumArrayInfo->C_array = 0;

	return (0);
}
