#include <Python.h>
#include <stdio.h>

/* 

This module exists as a work-around for a limitation in the Python-2.2 Buffer
Object: while the buffer object itself is always double aligned, the buffer
*storage* the object contains may not be.  The solution here is to create a
"memory" object which features double aligned storage, then create a buffer
object which references it.  This is necessary because destruction of a buffer
object only frees the object itself, and the object may not contain the
storage; freeing a buffer object will however DECREF any object it refers to.

This solution should enable numarray to work with Python >= 2.0.

*/

staticforward PyTypeObject MemoryType;

static PyObject *memoryError;

typedef struct {
	PyObject_VAR_HEAD
	char   *ptr;
	double pad[1];
} MemoryObject;

static PyObject *
new_memory(PyObject* self, PyObject* args)
{
	MemoryObject *memory;
	long ptr;
	long size;
	if (!PyArg_ParseTuple(args,"i", &size)) 
		return NULL;
	memory = PyObject_NewVar(MemoryObject, &MemoryType, size);
	/* Look past the padding */
	ptr = (long) (memory + 1);
	/* Truncate misaligned pointers into the padding. */
	memory->ptr = (char *) ((ptr/sizeof(double)) * sizeof(double));
	return (PyObject *) memory;
}

static PyObject *
memory_buffer(PyObject *self, PyObject *args)
{
	PyObject *memory = new_memory(self, args);
	if (memory) {
		PyObject *buffer = PyBuffer_FromReadWriteObject( memory, 
					  0, Py_END_OF_BUFFER);
		Py_DECREF(memory);  /* the buffer now "owns" the memory. */
		return buffer;
	} else {
		return NULL;
	}
}

static PyObject *
writeable_buffer(PyObject *self, PyObject *args)
{
  PyObject *ob;
  int offset = 0;
  int size = Py_END_OF_BUFFER;
  
  if ( !PyArg_ParseTuple(args, "O|ii:writeable_buffer", &ob, &offset, &size) )
    return NULL;
  return PyBuffer_FromReadWriteObject(ob, offset, size);
}

/* file_readinto is a slightly modified (fixed) copy of the same method
from python's fileobject.c modified to be a function rather than a type method.

This file_readinto fixes a bug exposed on Tru64 / OSF1 / v5.1 by declaring
the variable ntodo as an "int" rather than as "size_t".  
*/
static PyObject *
file_readinto(PyObject *self, PyObject *args)
{
	char *ptr;
	int   ntodo;
	size_t ndone, nnow;
	PyObject *fo;
	FILE *f;
	
	if (!PyArg_ParseTuple(args, "Ow#", &fo, &ptr, &ntodo))
		return NULL;
	
	if (!PyFile_Check(fo))
		return PyErr_Format(memoryError, 
			    "file_readinto(file, buffer) bad parameter list");
	else
		f = PyFile_AsFile(fo);

	ndone = 0;
	while (ntodo > 0) {
		Py_BEGIN_ALLOW_THREADS
		errno = 0;
		nnow = fread(ptr+ndone, 1, ntodo, f);
		Py_END_ALLOW_THREADS
		if (nnow == 0) {
			if (!ferror(f))
				break;
			PyErr_SetFromErrno(PyExc_IOError);
			clearerr(f);
			return NULL;
		}
		ndone += nnow;
		ntodo -= nnow;
	}
	return PyInt_FromLong((long)ndone);
}

static void
memory_dealloc(PyObject* self)
{
	PyObject_Del(self);
}

static PyObject * 
memory_repr(PyObject *self)
{
	MemoryObject *me = (MemoryObject *) self;
	char buffer[128];
	sprintf(buffer, 
		"<memory at %08lx with size:%d held by object at %08lx>",
		(long) me->ptr, me->ob_size, (long) self);
	return PyString_FromString(buffer);
}

/* Buffer methods */
static int
memory_getbuf(MemoryObject *self, int idx, void **pp)
{
	if ( idx != 0 ) {
		PyErr_SetString(memoryError,
				"memory objects only support one segment");
		return -1;
	}
	*pp = self->ptr;
	return self->ob_size;
}

static int
memory_getsegcount(MemoryObject *self, int *lenp)
{
	if ( lenp )
		*lenp = self->ob_size;
	return 1;
}

static long 
memory_length(MemoryObject *self)
{
	return self->ob_size;
}

static PySequenceMethods memory_as_sequence = {
	(inquiry)memory_length, /*sq_length*/
	(binaryfunc)0, /*sq_concat*/
	(intargfunc)0, /*sq_repeat*/
	(intargfunc)0, /*sq_item*/
	(intintargfunc)0, /*sq_slice*/
	(intobjargproc)0, /*sq_ass_item*/
	(intintobjargproc)0, /*sq_ass_slice*/
};

static PyBufferProcs memory_as_buffer = {
	(getreadbufferproc)memory_getbuf,
	(getwritebufferproc)memory_getbuf,
	(getsegcountproc)memory_getsegcount,
	(getcharbufferproc)memory_getbuf,
};

static PyTypeObject MemoryType = {
    PyObject_HEAD_INIT(NULL)
    0,
    "Memory",
    sizeof(MemoryObject),
    1,                           /* per item cost */
    memory_dealloc,              /* tp_dealloc */
    0,                           /* tp_print */
    0,                           /* tp_getattr */
    0,                           /* tp_setattr */
    0,                           /* tp_compare */
    memory_repr,                 /* tp_repr */
    0,                           /* tp_as_number */
    &memory_as_sequence,         /* tp_as_sequence */
    0,                           /* tp_as_mapping */
    0,                           /* tp_hash */
    0,                           /* tp_call */
    0,                           /* tp_str */
    0,                           /* tp_getattro */
    0,                           /* tp_setattro */
    &memory_as_buffer,           /* tp_as_buffer */
    0,                           /* tp_flags */
    0                            /* tp_doc */
};

static PyMethodDef memory_methods[] = {
    {"new_memory", new_memory, METH_VARARGS,
     "Create a new Memory object."},
    {"memory_buffer", memory_buffer, METH_VARARGS,
     "Create a new buffer object based on a Memory object."},
    {"writeable_buffer", writeable_buffer, METH_VARARGS,
     "Create a writeable buffer object referencing another python object"},
    {"file_readinto", file_readinto, METH_VARARGS,
     "Stores a file into any python object which supplies a writeable buffer"},
    {NULL, NULL, 0, NULL}
};

DL_EXPORT(void)
initmemory(void) 
{
	PyObject *d, *m;
	MemoryType.ob_type = &PyType_Type;
	m = Py_InitModule("memory", memory_methods);
	d = PyModule_GetDict(m);
	memoryError = PyErr_NewException("memory.error", NULL, NULL);
	PyDict_SetItemString(d, "error", memoryError);
	PyDict_SetItemString(d, "MemoryType", (PyObject *) &MemoryType);
}
