static PyObject *pNumArrayModule;
static PyObject *pNumArrayMDict;
static PyObject *pNumArrayClass;
static PyObject *pNumArrayError;
static PyObject *pNumericTypesModule;
static PyObject *pNumericTypesMDict;
static PyObject *pNumericTypesTDict;

static PyObject *pNumType[nNumericType];

#define ELEM(x) (sizeof(x)/sizeof(x[0]))

typedef struct
{
	char *name;
	int typeno;
} NumericTypeNameMapping;

static NumericTypeNameMapping NumericTypeNameMap[] = {
	{"Bool", tBool},
	{"Int8", tInt8},
	{"UInt8", tUInt8},
	{"Int16", tInt16},
	{"UInt16", tUInt16},
	{"Int32", tInt32},
	{"Float32", tFloat32},
	{"Float64", tFloat64},
	{"Complex64", tComplex64},
	{"Complex128", tComplex128}
};


/* Look up the NumericType which corresponds to typename */

static int NameToTypeNo(char *typename)
{
	int i;
	for(i=0; i<ELEM(NumericTypeNameMap); i++)
		if (!strcmp(typename, NumericTypeNameMap[i].name))
			return NumericTypeNameMap[i].typeno;
	return -1;
}

/* Convert NumericType 'typeno' into the string of the type's name. */

static char *TypeNoToName(int typeno) {
	int i;
	for(i=0; i<ELEM(NumericTypeNameMap); i++)
		if (typeno == NumericTypeNameMap[i].typeno)
			return NumericTypeNameMap[i].name;
	return 0;
}

/* Initialize this module. */

static void
NA_Init(void)
{
	int i;

	pNumArrayModule = PyImport_ImportModule("numarray");
	if (!pNumArrayModule)
		Py_FatalError("Error importing libnumarray.");
	pNumericTypesModule = PyImport_ImportModule("numerictypes");
	if (!pNumericTypesModule)
		Py_FatalError("Error importing numerictypes.");
	pNumArrayMDict = PyModule_GetDict(pNumArrayModule);
	if (!pNumArrayMDict)
		Py_FatalError("Error getting Numarray module dictionary.");
	Py_INCREF(pNumArrayMDict);
	pNumArrayClass = PyDict_GetItemString(pNumArrayMDict, "NumArray");
	if (!pNumArrayClass)
		Py_FatalError("Error getting NumArray class.");
	Py_INCREF(pNumArrayClass);
	pNumArrayError = PyDict_GetItemString(pNumArrayMDict, "Error");
	if (!pNumArrayError)
		Py_FatalError("Error getting Numarray.Error.");
	Py_INCREF(pNumArrayError);
	pNumericTypesMDict = PyModule_GetDict(pNumericTypesModule);
	if (!pNumericTypesMDict)
		Py_FatalError("Error getting numerictypes module dict.");
	Py_INCREF(pNumArrayMDict);
	pNumericTypesTDict = PyDict_GetItemString(pNumericTypesMDict,
						  "typeDict");
	if (!pNumericTypesTDict)
		Py_FatalError("Error getting numerictypes typeDict.");
	Py_INCREF(pNumericTypesTDict);

	for(i=0; i<ELEM(pNumType); i++) {
		PyObject *typeobj = PyDict_GetItemString(pNumericTypesTDict,
							 TypeNoToName(i));
		if (typeobj)
			pNumType[i] = typeobj;
		else
			Py_FatalError("Error getting type object.");
	}
}


/* Finalize this module. */

static void
NA_Done(void)
{
	int i;
	Py_DECREF(pNumArrayModule);
	Py_DECREF(pNumArrayClass);
	Py_DECREF(pNumArrayMDict);
	Py_DECREF(pNumArrayError);
	Py_DECREF(pNumericTypesModule);
	Py_DECREF(pNumericTypesMDict);
	Py_DECREF(pNumericTypesTDict);
	for(i=0; i<ELEM(pNumType); i++) {
		Py_DECREF(pNumType[i]);
	}
}

static PyObject *
tupleFromLongs(int len, long *Longs)
{
	int i;
	PyObject *intTuple = PyTuple_New(len);
	if (!intTuple)
		return NULL;
	for(i=0; i<len; i++)
		PyTuple_SET_ITEM( intTuple, i, PyInt_FromLong(Longs[i]));
	return intTuple;
}


/* count the number of elements in the given shape array. */
static int
nelements(int ndim, long *shape)
{
	int i, n;
	for(i=0, n=1; i<ndim; i++)
		n *= shape[i];
	return n;
}

/* Given a Numarray type object, return the size of the type in bytes. */
static long
sizeofType(PyObject *typeObj)
{
	PyObject *sizeInt = PyObject_GetAttrString(typeObj, "bytes");
	if (typeObj && sizeInt)	{
		long size = PyInt_AsLong(sizeInt);
		Py_DECREF(sizeInt);
		return size;
	} else {
		PyErr_Format(pNumArrayError,
			     "Error fetching .bytes from type");
		return -1;
	}
}

/* Count the number of elements in a 1D static array. */
#define ELEM(x) (sizeof(x)/sizeof(x[0]))

/* Create a new numarray specifying all attribute values. */
static PyObject *
NA_NewAll(int ndim, long *shape, NumericType type, void *buffer,
	  int byteoffset, int bytestride, 
	  int byteswap, int aligned, int writeable)
{
	PyObject *argsTuple, *shapeTuple, *typeObject, *bufferObject,
		*bytestrideObject;
	
	if (ndim > MAXDIM)
		return NULL;

	shapeTuple = tupleFromLongs(ndim, shape);
	if (!shapeTuple)
		return PyErr_Format(pNumArrayError, 
				    "Error creating shape tuple...");
	
	typeObject = pNumType[type];
	
	if (buffer) {
		int size = nelements(ndim, shape);
		if (bytestride) {
			size *= bytestride;
		} else {
			long typesize = sizeofType(typeObject);
			if ( typesize > 0)
				size *= sizeofType(typeObject);
			else
				return NULL;
		}
		if (writeable)
			bufferObject = PyBuffer_FromReadWriteMemory(buffer,
								    size);
		else
			bufferObject = PyBuffer_FromMemory(buffer, size);
	} else {
		Py_INCREF(Py_None);
		bufferObject = Py_None;
	}

	if (bytestride)	{
		bytestrideObject = PyInt_FromLong(bytestride);
		if (!bytestrideObject)
			return PyErr_Format(pNumArrayError, 
					    "Error making bytestride object.");
	}
	else {
		Py_INCREF(Py_None);
		bytestrideObject = Py_None;
	}

	argsTuple = Py_BuildValue("OOOlOii", shapeTuple, typeObject, 
				  bufferObject, byteoffset, bytestrideObject, 
				  byteswap, aligned);

	if (!argsTuple) {
		return PyErr_Format(pNumArrayError, 
				    "Error creating args tuple.");
	} else {
		return PyObject_CallObject(pNumArrayClass, argsTuple);
	}

	Py_DECREF(argsTuple);
}

/* Create a new numarray which is initially a C_array, or which
references a C_array: aligned, !byteswapped, contiguous, ... */

static PyObject *
NA_New(void *buffer, NumericType type, int ndim, ...)
{
	int i;
	long shape[MAXDIM];
	va_list ap;
	va_start(ap, ndim);
	for(i=0; i<ndim; i++)
		shape[i] = va_arg(ap, int);
	va_end(ap);
	return NA_NewAll(ndim, shape, type, buffer, 0, 0, 0, 1, 1);
}


/* Create a new array based on the information in the NDInfo*.  The new
array does not copy the data. */

static PyObject *
NA_FromInfo(NDInfo *info)
{
	return NA_NewAll(info->ndim, info->shape, 
			 NameToTypeNo(info->type), info->data,
			 info->byteoffset, info->bytestride, info->byteswap,
			 info->aligned, info->writeable);
}

/* Create a 1D numarray of the given NumericType from a C long. */

static PyObject* 
NA_FromLong(long l, NumericType type)
{
	long shape = 1;
	PyObject *a = NA_NewAll(1, &shape, type, NULL, 0, 0, 0, 1, 1);
	PyObject *lo = PyInt_FromLong(l);
	PyObject_SetItem(a, 0, lo);
	return a;
}

/* Create a 1D numarray of the given NumericType from a C double. */

static PyObject* 
NA_FromDouble(double d, NumericType type) 
{
	long shape = 1;
	PyObject *a = NA_NewAll(1, &shape, type, NULL, 0, 0, 0, 1, 1);
	PyObject *fo = PyFloat_FromDouble(d);
	PyObject_SetItem(a, 0, fo);
	return a;
}

/* Create the array [1] */

static PyObject*
NA_One()
{
	return NA_FromLong(1L, tInt32);
}

/* Create the array [0] */

static PyObject*
NA_Zero()
{
	return NA_FromLong(0L, tInt32);
}

/* getArray creates a new array of type 't' from the given array 'a'
using the specified 'method', probably 'new' or 'astype'. */

static PyObject *
getArray(PyObject *a, NDInfo *ainfo, NumericType t, char *method)
{
	char *name = TypeNoToName(t);
	PyObject *b;
	if (!name) 
		return PyErr_Format(pNumArrayError, "getArray: error getting name for type: %d.", t);
	b = PyObject_CallMethod(a, method, "s", name);
	if (!b) 
		return PyErr_Format(pNumArrayError, "getArray: error getting type '%s' using method '%s'.", name, method);
	if (getNumInfo(b, ainfo))
		return PyErr_Format(pNumArrayError, "getArray: error getting array info.");
	return b;
}

/* NA_ConvertInput is the main input conversion routine.  NA_ConvertInput
converts input array 'a' as necessary to NumericType 't' also guaranteeing
that either 'a' or the converted result is contigous, aligned, and not
byteswapped. NA_ConvertInput returns a pointer to a Numarray for which
C_array should be 1, and fills in 'ainfo' with the array information of
the return value.  The return value provides a means to later deallocate any
temporary array created by NA_ConvertInput, while 'ainfo' provides direct
access to the array's "metadata" from C.  Since the reference count of the
input array 'a' is incremented when it is directly useable by C, the return
value (either 'a' or a temporary) should always be passed to Py_DECREF by
the calller.
*/

static PyObject*
NA_ConvertInput(PyObject *a, NDInfo *ainfo, NumericType t)
{
	if (getNumInfo(a, ainfo))
		return PyErr_Format(pNumArrayError, 
				    "Numarray: error getting array info.");
	if (ainfo->C_array && NameToTypeNo(ainfo->type)==t) {
		Py_INCREF(a);
		return a;
	} else {
		return getArray(a, ainfo, t, "astype");
	}
} 

/* NA_ShadowOutput creates a C-usable temporary array similar to 'a' but of
type 't' as necessary.  If 'a' is already C-useable and of type 't', then 'a'
is returned.  In either case, 'ainfo' is filled in with the array information
for the return value.
*/

static PyObject *
NA_ShadowOutput(PyObject *a, NDInfo *ainfo, NumericType t)
{
	if (getNumInfo(a, ainfo))
		return PyErr_Format(pNumArrayError, 
				    "NA_ShadowOutput: error getting array info.");
	if (ainfo->C_array && NameToTypeNo(ainfo->type)==t) {
		return a;
	} else {
		return getArray(a, ainfo, t, "new");
	}
}

/* NA_ShadowOptionalOutput works like NA_ShadowOutput, but handles the case
where the output array 'optional' is omitted entirely at the python level,
resulting in 'optional'==NULL.  When 'optional' is NULL, the return value is
cloned (but with NumericType 't') from 'master', typically an input array with
the same shape as the output array.
*/

static PyObject *
NA_ShadowOptionalOutput(PyObject *optional, NDInfo *oinfo, 
			NumericType t, PyObject *master)
{
	if (!optional)
		return getArray(master, oinfo, t, "new");
	else
		return NA_ShadowOutput(optional, oinfo, t);
}

/* NA_ConvertOuput copies the contents of a 'shadow' array back into
the original un-usable array 'out' and deallocates 'shadow'.  If 'out'
was directly usable, no copy is performed because 'shadow' should
already be identical to 'out'.  
*/

static int
NA_ConvertOutput(PyObject *out, PyObject *shadow) {
	if (out != shadow) {
		if (!PyObject_CallMethod(out, "_copyFrom", "O", shadow)) {
			PyErr_Format(pNumArrayError,
				"NA_ConvertOutput: Error copying from shadow to output array.");
			return -1;
		}
		Py_DECREF(shadow);
	}
	return 0;
}

/* NA_ReturnShadowedOptionalOutput handles returning a possibly
unspecified output array.  If the array 'out' was specified on the
original call to the Python wrapper function, then the contents of any
'shadow' array are copied into 'out' as required.  The function then
returns Py_None.  If no output was specified in the original call,
then the 'shadow' array *becomes* the output and is returned.  This
results in extension functions which return Py_None when you specify
an output array, and return an array value otherwise.  These functions
also correctly handle data typing, alignment, byteswapping, and contiguity
issues */

static PyObject*
NA_ReturnShadowedOptionalOutput(PyObject *out, PyObject *shadow)
{
	if (out) { /* Since output array was specified, convert and return
		      nothing; the output array has the return value. */
		if (NA_ConvertOutput(out, shadow)) {
			return NULL;
		} else {
			Py_INCREF(Py_None);
			return Py_None;
		}
	} else { /* Since no output array was specified, the shadow *is*
		    the return value and no conversion is required. */
		return shadow;
	}
}

/* NA_ShapeMisMatch returns 1 if the given arrays (specified by NDInfo*) 'a'
 and 'b' have differing dimensionalities or shapes.  Otherwise it returns 0.
*/

static int
NA_ShapeMisMatch(NDInfo *a, NDInfo *b)
{
	if (a->ndim != b->ndim)
		return 1;
	else {
		int i;
		for(i=0; i<a->ndim; i++)
			if (a->shape[i] != b->shape[i])
				return 1;
		return 0;
	}
}

